{-|
  Module : DPSH
  (c) Akimasa Morihata (morihata\@ipl.t.u-tokyo.ac.jp), 2009

  Library for obtaining optimal lists

  This library helps us to easily develop efficient program 
  for obtaining optimal lists. 
  Embedded RULES pragma will improve written programs automatically.

  Its details are explained in the following paper:

    * Akimasa Morihata: 
      /A Short Cut to Optimal Sequences/.
       In Proc. the Seventh Asian Symposium on Programming 
       Languages and Systems (APLAS 2009). Accepted.
-}

{-# OPTIONS_GHC 
 -fglasgow-exts  -fenable-rewrite-rules
 -O -fvia-C -XNoMonomorphismRestriction
  #-}

module DPSH 
    ( -- * Sequence generators
      gen, 
      inits, tails, segs, subsequences, permutations,
      interleave, marking, markingBy, 
      -- * Filtering operations
      constraint, always, 
      -- * Maximization/minimization operations
      minBySum, minByMapSum, minByAccumSum,
      maxBySum, maxByMapSum, maxByAccumSum,
      minByLexico, minByMapLexico, minByAccumLexico,
      maxByLexico, maxByMapLexico, maxByAccumLexico)
where

import Debug.Trace

import Data.List hiding (inits,tails,subsequences,permutations) 
import Data.Ord
import Data.Array
import Data.Bits
import qualified Data.Map as Map

type Set a = [a]


{- ******** generators ******** -}

-- | @ gen g = g (++) (\\a->[a]) [] (\\a->map (a:)) (:) [] @
-- 
-- supply constructors to the functions of the polymorphic type and makes a build form.
gen :: forall c. Eq c => 
       (forall a b.
        (b->b->b)-> -- union
        (a->b)->    -- singleton
        b->         -- empty set
        (c->b->b)-> -- extend
        (c->a->a)-> -- cons
        a->         -- nil
        b) -> [[c]]
gen g = g (union) wrap [] extend (:) []
{-# INLINE [0] gen #-}


{- ******** useful list-generating computation patterns ******** -}

-- | identical to List.inits
inits :: forall a. Eq a => [a] -> [[a]]
inits x = gen (initsB x)
{-# INLINE inits #-}
initsB x u s e ex c n  = aux x 
    where aux [] = s n
          aux (a:x) = s n `u` ex a (aux x)
{-# INLINE initsB #-}

-- | identical to List.tails
tails :: forall a. Eq a => [a] -> [[a]]
tails x = gen (tailsB x)
{-# INLINE tails #-}
tailsB x u s e ex c n = fst $ aux x
    where aux [] = (s n, n)
          aux (a:x) = let (r,l) = aux x
                          l' = c a l
                      in (s l' `u` r, l')
{-# INLINE tailsB #-}

-- | generates all consecutive subsequences
segs :: forall a. Eq a => [a] -> [[a]]
segs x = gen (segsB x)
{-# INLINE segs #-}
segsB x u s e ex c n = fst $ aux x 
    where aux [] = (s n, s n)
          aux (a:x) = case aux x of 
                        (m,i) -> let i' = ex a i
                                 in (m `u` i', s n `u` i')
{-# INLINE segsB #-}

-- | identical to Data.List.subsequences
subsequences :: forall a. Eq a => [a] -> [[a]]
subsequences x = gen (subseqB x)
{-# INLINE subsequences #-}
subseqB x u s e ex c n = aux x
    where aux [] =  s n
          aux (a:x) = let r = aux x in ex a r `u` r 
{-# INLINE subseqB #-}


-- | generate all markings by Left or Right
marking :: forall a. Eq a => [a] -> [[Either a a]]
marking x = gen (markB x)
{-# INLINE marking #-}
markB x u s e ex c n = aux x 
    where aux [] = s n
          aux (a:x) = let r = aux x in ex (Left a) r `u` ex (Right a) r
{-# INLINE markB #-}

-- | generate all markings by the given list of functions 
markingBy :: forall a b. Eq b => [a->b] -> [a] -> [[b]]
markingBy mks x = gen (markByB mks x)
{-# INLINE markingBy #-}
markByB mks x u s e ex c n = aux mks x
    where aux mks [] = s n
          aux mks (a:x) = let r = aux mks x 
                          in aux2 mks a r 
          aux2 [] a r = e
          aux2 (f:mks) a r = ex (f a) r `u` aux2 mks a r
{-# INLINE markByB #-}

-- | identical to Data.List.permutations
permutations :: forall a. Eq a => [a] -> [[a]]
permutations x = gen permB
  where 
    len = length x
    arr = listArray (1,len) x
    permB u s e ex c n = 
        let aux 0 h = case Map.elems h of [xs] -> xs
            aux (l+1) h = aux l $ Map.foldWithKey f Map.empty h
            f b x m = foldr (\k r->if not (testBit b k) 
                                   then Map.insertWith u (setBit b k) 
                                            (ex (arr!k) x) r 
                                   else r) m [1..len]
        in aux len (Map.singleton (0::Int) (s n))
    {-# INLINE permB #-}
{-# INLINE permutations #-}

-- | generate all lists that are \"interleaving\" of the given two lists
interleave :: forall a. Eq a => [a] -> [a] -> [[a]]
interleave x y = gen intrB 
  where 
    lx = length x
    ly = length y
    arx = listArray (1,lx) x
    ary = listArray (1,ly) y
    intrB u s e ex c n = 
        let aux 0 h = case Map.elems h of [xs] -> xs
            aux (k+1) h = aux k $ Map.foldWithKey f Map.empty h
            f (a,b) z m 
                | a==0 = Map.insertWith u (a,b-1) (ex (ary!b) z) m
                | b==0 = Map.insertWith u (a-1,b) (ex (arx!a) z) m
                | otherwise = Map.insertWith u (a-1,b) (ex (arx!a) z) $ 
                              Map.insertWith u (a,b-1) (ex (ary!b) z) m
        in aux (lx+ly) (Map.singleton (lx,ly) (s n))
    {-# INLINE intrB #-}
{-# INLINE interleave #-}


{- ******** filterings ******** -}

-- | @ constraint p f e = filter (p . foldr f e) @
constraint :: forall a b. (Eq b, Ord a) => 
              (a->Bool) -> (b->a->a) -> a -> [[b]] -> [[b]]
constraint p f e = filter (p.foldr f e)
{-# INLINE [0] constraint #-}

-- | @ always p f e = filter (all p . scanr f e) @
always :: forall a b. (Eq b, Ord a) => 
          (a->Bool) -> (b->a->a) -> a -> [[b]] -> [[b]]
always p f e = filter (all p . scanr f e)
{-# INLINE [0] always #-}
{-# RULES "inline-always" [~1]
  forall p f e. 
  always p f e = 
    let cmp (t1,s1) (t2,s2) = case compare t1 t2 of 
                                      LT -> PLT
                                      GT -> PGT
                                      EQ -> if s1 == s2 then PEQ else PNEQ
    in maxIR (\a (r,s)->let s' = f a s in (r && p s', s')) (p e,e) 
                       cmp (always' p f e)
  #-}

{- ******** minimizations and maximizations ******** -}

-- | @minBySum@ extracts the lists whose sum is minimum
minBySum :: forall a. (Num a, Ord a) => [[a]] -> [[a]]
minBySum = maxByMapSum negate
{-# INLINE minBySum #-}

-- | @minByMapSum f@ extracts the lists whose @ sum . map f @ is minimum
minByMapSum :: forall a b. (Eq a, Num b, Ord b) => (a->b) -> [[a]] -> [[a]]
minByMapSum f = maxByMapSum (negate.f)
{-# INLINE minByMapSum #-}

-- | @minByAccumSum f e@ extracts the lists whose @ sum . snd . mapAccumR f e @ is minimum
minByAccumSum :: forall a b c. (Eq a, Ord b, Num c, Ord c) => 
                 (b->a->(b,c)) -> b -> [[a]] -> [[a]]
minByAccumSum f e = maxByAccumSum (\s a->let (s',a') = f s a in (s',-a')) e
{-# INLINE minByAccumSum #-}

-- | @maxBySum@ extracts the lists whose sum is maximum
maxBySum :: forall a. (Num a, Ord a) => [[a]] -> [[a]]
maxBySum = map snd . maximals pcmp . map (\a-> (sum a,a))
{-# INLINE [0] maxBySum #-}
{-# RULES "inline-maxSum" [~1] 
  maxBySum = maxIR (+) 0 pcmp id 
  #-}

-- | @maxByMapSum f@ extracts the lists whose @ sum . map f @ is maximum
maxByMapSum :: forall a b. (Eq a, Num b, Ord b) => (a->b) -> [[a]] -> [[a]]
maxByMapSum f = map snd . maximals pcmp . map (\a-> (sum $ map f a,a))
{-# INLINE [0] maxByMapSum #-}
{-# RULES "inline-maxMapSum" [~1] 
  forall f. maxByMapSum f = maxIR (\a r ->f a + r) 0 pcmp id 
  #-}

-- | @maxByAccumSum f e@ extracts the lists whose @ sum . snd . mapAccumR f e@ is maximum
maxByAccumSum :: forall a b c. (Eq a, Ord b, Num c, Ord c) => 
                 (b->a->(b,c)) -> b -> [[a]] -> [[a]]
maxByAccumSum f e 
    = map snd . maximals pcmp . map (\a->(sum $ snd $ mapAccumR f e a,a)) 
{-# INLINE [0] maxByAccumSum #-}
{-# RULES "inline-maxAccumSum" [~1] 
  forall f e. 
  maxByAccumSum f e 
    = let cmp (s1,v1) (s2,v2) = if s1 == s2 then pcmp v1 v2 else PNEQ
      in maxIR (\a (s,r)->let (s',a') = f s a in (s',a'+r)) (e,0) 
            cmp (maxByAccumSum' f e)
  #-}

-- | @minByLexico@ extracts the lists that are the minimum on lexicographic ordering
minByLexico :: forall a. (Ord a) => [[a]] -> [[a]]
minByLexico = map snd . maximals pccmp . map (\a-> (a,a))
{-# INLINE [0] minByLexico #-}
{-# RULES "inline-minLexico" [~1] 
  minByLexico = maxIR (:) [] pccmp id 
  #-}

-- | @minByMapLexico@ extracts the lists whose @ map f @ are the minimum on lexicographic ordering
minByMapLexico :: forall a b. (Eq a, Ord b) => (a->b) -> [[a]] -> [[a]]
minByMapLexico f = map snd . maximals pccmp . map (\a-> (map f a,a))
{-# INLINE [0] minByMapLexico #-}
{-# RULES "inline-minMapLexico" [~1] 
  forall f. minByMapLexico f = maxIR (\a r ->f a:r) [] pccmp id 
  #-}

-- | @minByAccumLexico@ extracts the lists whose @ snd . mapAccumR f e @ are the minimum on lexicographic ordering
minByAccumLexico :: forall a b c. (Eq a, Ord b, Ord c) => 
                    (b->a->(b,c)) -> b -> [[a]] -> [[a]]
minByAccumLexico f e
    = map snd . maximals pccmp . map (\a->(snd $ mapAccumR f e a,a))
{-# INLINE [0] minByAccumLexico #-}
{-# RULES "inline-minAccumLexico" [~1] 
  forall f e. 
  minByAccumLexico f e 
    = let cmp (s1,v1) (s2,v2) = if s1 == s2 then pccmp v1 v2 else PNEQ
      in maxIR (\a (s,r)->let (s',a') = f s a in (s',a':r)) (e,[]) 
            pccmp (minByAccumLexico' f e)
  #-}

-- | @maxByLexico@ extracts the lists that are the maximum on lexicographic ordering
maxByLexico :: forall a. (Ord a) => [[a]] -> [[a]]
maxByLexico = map snd . maximals pcmp . map (\a-> (a,a))
{-# INLINE [0] maxByLexico #-}
{-# RULES "inline-maxLexico" [~1] 
  maxByLexico = maxIR (:) [] pcmp id 
  #-}

-- | @maxByMapLexico f@ extracts the lists whose @ map f @ are the maximum on lexicographic ordering
maxByMapLexico :: forall a b. (Eq a, Ord b) => (a->b) -> [[a]] -> [[a]]
maxByMapLexico f = map snd . maximals pcmp . map (\a-> (map f a,a))
{-# INLINE [0] maxByMapLexico #-}
{-# RULES "inline-maxMapLexico" [~1] 
  forall f. maxByMapLexico f = maxIR (\a r ->f a:r) [] pcmp id 
  #-}


-- | @maxByAccumLexico@ extracts the lists whose @ snd . mapAccumR f e @ are the maximum on lexicographic ordering
maxByAccumLexico :: forall a b c. (Eq a, Ord b, Ord c) => 
                    (b->a->(b,c)) -> b -> [[a]] -> [[a]]
maxByAccumLexico f e 
    = map snd . maximals pcmp . map (\a-> (snd $ mapAccumR f e a,a))
{-# INLINE [0] maxByAccumLexico #-}
{-# RULES "inline-maxAccumLexico" [~1] 
  forall f e. 
  maxByAccumLexico f e 
    = let cmp (s1,v1) (s2,v2) = if s1 == s2 then pcmp v1 v2 else PNEQ
      in maxIR (\a (s,r)-> let (s',a') = f s a in (s',a':r)) (e,[]) 
               cmp (maxByAccumLexico' f e)
  #-}



{- ******** definitions of orders ******** -}

data POrdering = PGT | PLT | PEQ | PNEQ

pcmp a b = case compare a b of GT -> PGT 
                               LT -> PLT
                               EQ -> PEQ
{-# INLINE pcmp #-}
pccmp a b = case compare a b of GT -> PLT 
                                LT -> PGT
                                EQ -> PEQ
{-# INLINE pccmp #-}

{- ******** core representation ******** -}

maxIR :: forall a b. (Eq a, Ord b) => 
         (a->b->b) -> b -> (b->b->POrdering) ->
                      (Set [a]->Set [a])-> Set [a] -> Set [a]
maxIR f e cmps k x = k $ map snd $ maximals cmps $
                     map (\a->(foldr f e a, a)) x
{-# INLINE [0] maxIR #-}


{- ******** rewrite rules ******** -}

newtype Sorted a b = C (Set (a,[b]))


{-# RULES "shortcut fusion for DP" [1] 
  forall f e cmp k
         (g::forall a b. (b->b->b)->(a->b)->b->(c->b->b)->(c->a->a)->a->b).
  maxIR f e cmp k (gen g) 
    = let con' a (v,x) = (f a v,map (a:) x)
          con a (v,x) = (f a v, a:x)
          up x y = maximals' cmp (x `merge` y)
          {-# INLINE up #-}
          ext a (C x) = maximals' cmp (sortC $ map (con' a) x)
          {-# INLINE ext #-}
      in case g up (\(v,a)-> C [(v,[a])]) (C []) ext con (e,[]) of 
         C rs -> k $ concatMap snd rs
  #-}

{-# RULES "constraint-constraint"
  forall p1 f1 e1 p2 f2 e2.
  constraint p1 f1 e1 . constraint p2 f2 e2
    = constraint (\(s,t)->p1 s && p2 t) (\a (s,t)->(f1 a s,f2 a t)) (e1,e2)
  #-}

{-# RULES "constraint-constraint-eta"
  forall p1 f1 e1 p2 f2 e2 x.
  constraint p1 f1 e1 (constraint p2 f2 e2 x) 
    = constraint (\(s,t)->p1 s && p2 t) (\a (s,t)->(f1 a s,f2 a t)) (e1,e2) x
  #-}

{-# RULES "max-constraint" [~1]
  forall f e cmp k p g z.
  maxIR f e cmp k . constraint p g z 
    = let cmp' (a1,b1) (a2,b2) = if a1 == a2 then cmp b1 b2 else PNEQ
      in maxIR (\a (s,v)->(g a s, f a v)) (z,e) cmp'
            (maxIR' f e cmp k . constraint p g z)
  #-}

{-# RULES "max-constraint-eta" [~1]
  forall f e cmp k p g z x.
  maxIR f e cmp k (constraint p g z x) 
    = let cmp' (a1,b1) (a2,b2) = if a1 == a2 then cmp b1 b2 else PNEQ
      in maxIR (\a (s,v)->(g a s, f a v)) (z,e) cmp'
            (maxIR' f e cmp k . constraint p g z) x
  #-}

{-# RULES "max-max" [~1]
  forall f1 e1 cmp1 k1 f2 e2 cmp2 k2.
  maxIR f1 e1 cmp1 k1 . maxIR f2 e2 cmp2 k2
    = let cmp' (a1,b1) (a2,b2) = case cmp2 a1 a2 of PEQ -> cmp1 b1 b2; r -> r
      in maxIR (\a (r2,r1) ->(f2 a r2,f1 a r1)) (e2,e1) cmp'
            (maxIR' f1 e1 cmp1 k1 . k2)
  #-}


{-# RULES "max-max-eta" [~1]
  forall f1 e1 cmp1 k1 f2 e2 cmp2 k2 x.
  maxIR f1 e1 cmp1 k1 (maxIR f2 e2 cmp2 k2 x)
    = let cmp' (a1,b1) (a2,b2) = case cmp2 a1 a2 of PEQ -> cmp1 b1 b2; r -> r
      in maxIR (\a (r2,r1) ->(f2 a r2,f1 a r1)) (e2,e1) cmp'
            (maxIR' f1 e1 cmp1 k1 . k2) x
  #-}


{- ******** supplemental rules ******** -}

{-# RULES "max-cancel1"
  forall cmp x y.
  maximals' cmp (maximals' cmp x `merge` y) = maximals' cmp (x `merge` y)
  #-}

{-# RULES "max-cancel2"
  forall cmp x y.
  maximals' cmp (x `merge` maximals' cmp y) = maximals' cmp (x `merge` y)
  #-}

{-# RULES "max-cancel3"
  forall cmp x y.
  maximals' cmp (maximals' cmp x `merge` maximals' cmp y) 
    = maximals' cmp (x `merge` y)
  #-}


{- ******** auxiliary functions ******** -}

extend :: forall c. c -> Set [c] -> Set [c]
extend a = map (a:)

wrap :: forall a. a -> Set a
wrap a = [a]


maximals :: forall a b. (Eq a, Ord b) 
            => (b->b->POrdering) -> Set (b,a) -> Set (b,a) 
maximals f [] = []
maximals f x  = case (maximals' f $ sortC $ map (\(u,s)->(u,[s])) x) of 
                  C r -> concatMap (\(u,ss)->[(u,s)| s<-ss]) r 

maximals' :: forall a b. (Eq a, Ord b) 
             => (b->b->POrdering) -> Sorted b a -> Sorted b a
maximals' f (C []) = C []
maximals' f y@(C [r]) = y
maximals' f (C ((u,ss):y))  = case aux u ss y [] of 
                                [r] -> C [r]
                                (u',ss'):y' -> C $ aux u' ss' y' []
                                    -- in traceShow (length x,length r) r
    where aux u ss [] h = (u,ss):h
          aux u ss ((v,ts):x) h = 
              case f u v of PGT -> aux u ss x h
                            PLT -> aux v ts x h
                            PEQ -> aux u (myNub ts ss) x h
                            PNEQ -> aux v ts x ((u,ss):h)
          myNub [] y = y
          myNub (a:x) y | a `elem` y = myNub x y
                        | otherwise = a : myNub x y
{-# INLINE [0] maximals' #-}

merge :: forall a b. Ord b 
         => Sorted b a -> Sorted b a -> Sorted b a
merge (C x) (C y) = C $ mergeMain x y

mergeMain [] ys = ys
mergeMain xs [] = xs
mergeMain (x@(u,ss):xs) (y@(v,ts):ys)
    = case compare u v of GT -> y : mergeMain (x:xs) ys
                          _  -> x : mergeMain xs (y:ys)
{-# INLINE [0] merge #-}

sortC [] = C []
sortC x@[a] = C x
sortC x = case segmentify x of
            (b,bs,xs) -> C $ aux ((b:bs):xs)
    where segmentify [a] = (a,[],[])
          segmentify (a@(av,al):x) = 
              case segmentify x of 
                (b@(bv,bl),bs,xs) -> 
                    case compare av bv of 
                      GT -> (a,[],(b:bs):xs)
                      _ -> (a,b:bs,xs)
          aux [x] = x
          aux xs = aux $ aux2 xs
          aux2 [] = []
          aux2 xs@[x] = xs
          aux2 (x:y:xs) = mergeMain x y : aux2 xs

always' = always
{-# INLINE [0] always' #-}

maxByAccumSum' = maxByAccumSum
{-# INLINE [0] maxByAccumSum' #-}

maxByAccumLexico' = maxByAccumLexico
{-# INLINE [0] maxByAccumLexico' #-}


minByAccumLexico' = minByAccumLexico
{-# INLINE [0] minByAccumLexico' #-}

maxIR' = maxIR
{-# INLINE [0] maxIR' #-}

